Imports System.Device.Location
Imports System.IO.IsolatedStorage
Imports System.Text
Imports System.Windows
Imports System.Windows.Controls
Imports System.Windows.Media
Imports System.Windows.Media.Imaging
Imports System.Windows.Navigation
Imports Microsoft.Devices.Sensors
Imports Microsoft.Phone.Controls
Imports Microsoft.Phone.Shell
Imports Microsoft.Xna.Framework
Imports RotatingMapTiles.ImageryService

Namespace RotatingMapTiles
	Partial Public Class MainPage
		Inherits PhoneApplicationPage

		Private Const BITRES As Integer = 29

		Private geoWatcher As New GeoCoordinateWatcher(GeoPositionAccuracy.High)
		Private motion As Motion
		Private strBuilder As New StringBuilder

		Private roadUriTemplate, aerialUriTemplate As String
		Private mapStyle As MapStyle
		Private zoomLevel As Integer
		Private minimumLevel, maximumLevel As Integer
		Private integerLongitude As Integer = -1
		Private integerLatitude As Integer = -1

		Public Sub New()
			InitializeComponent()

			' Fix that annoying ApplicationBar bug
			minusAppBarIconButton = TryCast(Me.ApplicationBar.Buttons(1), ApplicationBarIconButton)
			plusAppBarIconButton = TryCast(Me.ApplicationBar.Buttons(2), ApplicationBarIconButton)

			' Create 25 Image elements for a 5x5 array
			For i  = 0 To 24
				Dim image As Image = New Image With {.Source = New BitmapImage, .Stretch = Stretch.None}
				imageCanvas.Children.Add(image)
			Next i

			' Initialize the Motion sensor
			If Motion.IsSupported Then
				motion = New Motion
				motion.TimeBetweenUpdates = TimeSpan.FromMilliseconds(33)
				AddHandler motion.CurrentValueChanged, AddressOf OnMotionCurrentValueChanged
			End If

			' Two more event handlers
			AddHandler geoWatcher.PositionChanged, AddressOf OnGeoWatcherPositionChanged
			AddHandler Loaded, AddressOf OnMainPageLoaded
		End Sub

		Protected Overrides Sub OnNavigatedTo(ByVal args As NavigationEventArgs)
			' Get saved information
			Dim settings As IsolatedStorageSettings = IsolatedStorageSettings.ApplicationSettings

			If Not settings.TryGetValue(Of Integer)("ZoomLevel", zoomLevel) Then
				zoomLevel = 12
			End If

			If Not settings.TryGetValue(Of MapStyle)("MapStyle", mapStyle) Then
				mapStyle = MapStyle.Road
			End If

			' Start the Motion sensor GPS watcher
			If motion IsNot Nothing Then
				Try
					motion.Start()
				Catch
				End Try
			End If

			geoWatcher.Start()
			MyBase.OnNavigatedTo(args)
		End Sub

		Protected Overrides Sub OnNavigatedFrom(ByVal args As NavigationEventArgs)
			' Stop the Motion sensor and GPS watcher
			If motion IsNot Nothing Then
				motion.Stop()
			End If

			geoWatcher.Stop()

			' Save information
			Dim settings As IsolatedStorageSettings = IsolatedStorageSettings.ApplicationSettings
			settings("ZoomLevel") = zoomLevel
			settings("MapStyle") = mapStyle
			settings.Save()

			MyBase.OnNavigatedFrom(args)
		End Sub

		Private Sub OnMainPageLoaded(ByVal sender As Object, ByVal e As RoutedEventArgs)
			' Initialize the Bing Maps imagery service
			Dim imageryServiceClient As New ImageryServiceClient("BasicHttpBinding_IImageryService")
			AddHandler imageryServiceClient.GetImageryMetadataCompleted, AddressOf GetImageryMetadataCompleted

			' Make a request for the road metadata
			Dim request As ImageryMetadataRequest = New ImageryMetadataRequest With {.Credentials = New Credentials With {.ApplicationId = "credentials-key"}, .Style = MapStyle.Road}
			imageryServiceClient.GetImageryMetadataAsync(request, "road")

			' Make a request for the aerial metadata
			request.Style = MapStyle.Aerial
			imageryServiceClient.GetImageryMetadataAsync(request, "aerial")
		End Sub

		Private Sub GetImageryMetadataCompleted(ByVal sender As Object, ByVal args As GetImageryMetadataCompletedEventArgs)
			If (Not args.Cancelled) AndAlso args.Error Is Nothing Then
				' Get the "powered by" bitmap
				poweredByBitmap.UriSource = args.Result.BrandLogoUri
				poweredByDisplay.Visibility = Visibility.Visible

				' Get the range of map levels available
				Dim result As ImageryMetadataResult = args.Result.Results(0)
				minimumLevel = result.ZoomRange.From
				maximumLevel = result.ZoomRange.To

				' Get the URI and make some substitutions
				Dim uri As String = result.ImageUri
				uri = uri.Replace("{subdomain}", result.ImageUriSubdomains(0))
				uri = uri.Replace("&token={token}", "")
				uri = uri.Replace("{culture}", "en-US")

				If TryCast(args.UserState, String) = "road" Then
					roadUriTemplate = uri
				Else
					aerialUriTemplate = uri
				End If

				If roadUriTemplate IsNot Nothing AndAlso aerialUriTemplate IsNot Nothing Then
					RefreshDisplay()
				End If
			Else
				errorTextBlock.Text = "Cannot access Bing Maps: " & args.Error.Message
			End If
		End Sub

		Private Sub OnMotionCurrentValueChanged(ByVal sender As Object, ByVal args As SensorReadingEventArgs(Of MotionReading))
				' Get the rotation matrix & convert to horizontal coordinates
				' Set the azimuth angle to rotation transforms
			Me.Dispatcher.BeginInvoke(Sub()
				Dim matrix As Microsoft.Xna.Framework.Matrix = args.SensorReading.Attitude.RotationMatrix
				Dim vector As Vector3 = Vector3.Transform(Vector3.UnitY, matrix)
				Dim azimuth As Double = 180 * Math.Atan2(vector.X, vector.Y) / Math.PI
				imageCanvasRotate.Angle = azimuth
				northArrowRotate.Angle = azimuth
			End Sub)
		End Sub

		Private Sub OnGeoWatcherPositionChanged(ByVal sender As Object, ByVal args As GeoPositionChangedEventArgs(Of GeoCoordinate))
			' Make the location cirlc visible
			locationDisplay.Visibility = Visibility.Visible

			' Save the integer longitude and latitude
			Dim longitude As Double = args.Position.Location.Longitude
			Dim latitude As Double = args.Position.Location.Latitude

			Dim relativeLongitude As Double = (180 + longitude) / 360
            integerLongitude = CInt(relativeLongitude * (1 << BITRES))

			Dim sinTerm As Double = Math.Sin(Math.PI * latitude / 180)
			Dim relativeLatitude As Double = 0.5 - Math.Log((1 + sinTerm) / (1 - sinTerm)) / (4 * Math.PI)
            integerLatitude = CInt((relativeLatitude * (1 << BITRES)))

			RefreshDisplay()
		End Sub

		Private Sub OnToggleViewAppBarButtonClick(ByVal sender As Object, ByVal args As EventArgs)
			If mapStyle = MapStyle.Road Then
				mapStyle = MapStyle.Aerial
			Else
				mapStyle = MapStyle.Road
			End If

			RefreshDisplay()
		End Sub

		Private Sub OnZoomInAppBarButtonClick(ByVal sender As Object, ByVal args As EventArgs)
			zoomLevel += 1
			EnableZoomButtons()
			RefreshDisplay()
		End Sub

		Private Sub OnZoomOutAppBarButtonClick(ByVal sender As Object, ByVal args As EventArgs)
			zoomLevel -= 1
			EnableZoomButtons()
			RefreshDisplay()
		End Sub

		Private Sub EnableZoomButtons()
			minusAppBarIconButton.IsEnabled = zoomLevel > minimumLevel
			plusAppBarIconButton.IsEnabled = zoomLevel < maximumLevel
		End Sub

		Private Sub RefreshDisplay()
			If roadUriTemplate Is Nothing OrElse aerialUriTemplate Is Nothing Then
				Exit Sub
			End If

			If integerLongitude = -1 OrElse integerLatitude = -1 Then
				Exit Sub
			End If

			' Get coordinates and pixel offsets based on current zoom level
			Dim croppedLongitude As Integer = integerLongitude >> BITRES - zoomLevel
			Dim croppedLatitude As Integer = integerLatitude >> BITRES - zoomLevel
			Dim xPixelOffset As Integer = (integerLongitude >> BITRES - zoomLevel - 8) Mod 256
			Dim yPixelOffset As Integer = (integerLatitude >> BITRES - zoomLevel - 8) Mod 256

			' Prepare for the loop
			Dim uriTemplate As String = If(mapStyle = MapStyle.Road, roadUriTemplate, aerialUriTemplate)
			Dim index As Integer = 0
			Dim maxValue As Integer = (1 << zoomLevel) - 1

			' Loop through the 5x5 array of Image elements
			For row  = -2 To 2
				For col  = -2 To 2
					' Get the Image and BitmapImage
					Dim image As Image = TryCast(imageCanvas.Children(index), Image)
					Dim bitmap As BitmapImage = TryCast(image.Source, BitmapImage)
					index += 1

					' Check if we've gone beyond the bounds
					If croppedLongitude + col < 0 OrElse croppedLongitude + col > maxValue OrElse croppedLatitude + row < 0 OrElse croppedLatitude + row > maxValue Then
						bitmap.UriSource = Nothing
					Else
						' Calculate a quadkey and set URI to bitmap
						Dim longitude As Integer = croppedLongitude + col
						Dim latitude As Integer = croppedLatitude + row
						Dim strQuadkey As String = ToQuadKey(longitude, latitude, zoomLevel)
						Dim uri As String = uriTemplate.Replace("{quadkey}", strQuadkey)
						bitmap.UriSource = New Uri(uri)
					End If

					' Position the Image element
					Canvas.SetLeft(image, col * 256 - xPixelOffset)
					Canvas.SetTop(image, row * 256 - yPixelOffset)
				Next col
			Next row
		End Sub

		Private Function ToQuadKey(ByVal longitude As Integer, ByVal latitude As Integer, ByVal level As Integer) As String
			Dim quadkey As Long = 0
			Dim mask As Integer = 1 << (level - 1)

			For i  = 0 To level - 1
				quadkey <<= 2

                If (longitude And mask) <> 0 Then
                    quadkey = quadkey Or 1
                End If

				If (latitude And mask) <> 0 Then
					quadkey = quadkey Or 2
				End If

				mask >>= 1
			Next i

			strBuilder.Clear()

			For i  = 0 To level - 1
				strBuilder.Insert(0, (quadkey And 3).ToString)
				quadkey >>= 2
			Next i

			Return strBuilder.ToString
		End Function
	End Class
End Namespace
